__version__ = "0.1.2"
__author__ = "Forest Dussault"
__email__ = "forest.dussault@canada.ca"

DEPENDENCIES = [
    'skesa',
    'bbduk.sh',
    'bbmap.sh',
    'tadpole.sh',
    'pilon',
    'samtools'
]



#!/usr/bin/env python3

import os
import click
import logging

from pathlib import Path
from ProkaryoteAssembly.prokaryote_assemble import assembly_pipeline, basic_cleanup, total_cleanup
from ProkaryoteAssembly.accessories import print_version, convert_to_path

script = os.path.basename(__file__)
logger = logging.getLogger()
logging.basicConfig(
    format=f'\033[92m \033[1m {script}:\033[0m %(message)s ',
    level=logging.DEBUG)


@click.command()
@click.option('-i', '--input_dir',
              type=click.Path(exists=False),
              required=True,
              help='Directory containing all *.fastq.gz files to assemble.'
                   'NOTE: Files must be gzipped in order to be detected.',
              callback=convert_to_path)
@click.option('-o', '--out_dir',
              type=click.Path(exists=False),
              required=True,
              help='Root directory to store all output files.',
              callback=convert_to_path)
@click.option('-f', '--fwd_id',
              type=click.STRING,
              required=False,
              default="_R1",
              help='Pattern to detect forward reads. Defaults to "_R1".')
@click.option('-r', '--rev_id',
              type=click.STRING,
              required=False,
              default="_R2",
              help='Pattern to detect reverse reads. Defaults to "_R2".')
@click.option('-m', '--memory',
              type=click.STRING,
              required=False,
              default='4g',
              help='Memory to allocate to pilon call. Defaults to 4g (i.e. pilon -Xmx4g). May need to provide a large '
                   'amount of memory for large read sets/assemblies.')
@click.option('--cleanup',
              help='Specify this flag to delete all intermediary files except the resulting FASTA assembly.',
              default=False,
              required=False,
              is_flag=True)
@click.option('--version',
              help='Specify this flag to print the version and exit.',
              is_flag=True,
              is_eager=True,
              callback=print_version,
              expose_value=False)
def assemble_dir(input_dir, out_dir, fwd_id, rev_id, memory, cleanup):
    if out_dir.exists():
        logging.error(f"ERROR: Output directory {out_dir} already exists. Specify a directory that does not yet exist.")
        quit()

    out_dir.mkdir(parents=True)
    logging.info(f"Created output directory {out_dir}")
    logging.info(f"Finding FASTQ files in {input_dir}")
    fastq_file_list = retrieve_fastqgz(input_dir)

    logging.info(f"Finding Sample IDs")
    sample_ids = retrieve_sampleids(fastq_file_list)

    logging.info(f"Found {len(sample_ids)} unique sample IDs")

    for sample_id in sample_ids:
        r1, r2 = get_readpair(sample_id=sample_id, fastq_file_list=fastq_file_list,
                              forward_id=fwd_id, reverse_id=rev_id)
        sample_out_dir = out_dir / sample_id
        os.makedirs(sample_out_dir, exist_ok=True)
        assembly_pipeline(fwd_reads=r1, rev_reads=r2, out_dir=sample_out_dir, sample_id=sample_id, memory=memory)
        if cleanup:
            total_cleanup(input_dir=sample_out_dir)
        else:
            basic_cleanup(input_dir=sample_out_dir)
            logging.info(f"Assembly complete for {sample_id} ({out_dir})")
    logging.info("Script complete!")


def retrieve_fastqgz(directory: Path) -> [Path]:
    """
    :param directory: Path to folder containing output from MiSeq run
    :return: LIST of all .fastq.gz files in directory
    """
    fastq_file_list = list(directory.glob("*.f*q.gz"))
    return fastq_file_list


def retrieve_sampleids(fastq_file_list: [Path]) -> list:
    """
    :param fastq_file_list: List of fastq.gz filepaths generated by retrieve_fastqgz()
    :return: List of Sample IDs
    """
    # Iterate through all of the fastq files and grab the sampleID, append to list
    sample_id_list = list()
    for f in fastq_file_list:
        sample_id = f.name.split('_')[0]
        sample_id_list.append(sample_id)

    # Get unique sample IDs
    sample_id_list = list(set(sample_id_list))
    return sample_id_list


def get_readpair(sample_id: str, fastq_file_list: [Path], forward_id: str,
                 reverse_id: str) -> (list, None):
    """
    :param sample_id: String of sample ID
    :param fastq_file_list: List of fastq.gz file paths generated by retrieve_fastqgz()
    :param forward_id: ID indicating forward read in filename (e.g. _R1)
    :param reverse_id: ID indicating reverse read in filename (e.g. _R2)
    :return: the absolute filepaths of R1 and R2 for a given sample ID
    """
    r1, r2 = None, None
    for f in fastq_file_list:
        if sample_id in f.name:
            if forward_id in f.name:
                r1 = f
            elif reverse_id in f.name:
                r2 = f
    if r1 is not None and r2 is not None:
        return [r1, r2]
    else:
        logger.info('Could not pair {}'.format(sample_id))
        return None


if __name__ == "__main__":
    assemble_dir()



#!/usr/bin/env python3

import os
import click
import shutil
import logging
from pathlib import Path

from ProkaryoteAssembly.accessories import print_version, convert_to_path, run_subprocess, check_all_dependencies

script = os.path.basename(__file__)
logger = logging.getLogger()
logging.basicConfig(
    format=f'\033[92m \033[1m {script}:\033[0m %(message)s ',
    level=logging.INFO)


@click.command()
@click.option('-1', '--fwd_reads',
              type=click.Path(exists=True),
              required=True,
              help='Path to forward reads (R1) (gzipped FASTQ).',
              callback=convert_to_path)
@click.option('-2', '--rev_reads',
              type=click.Path(exists=True),
              required=True,
              help='Path to reverse reads (R2) (gzipped FASTQ).',
              callback=convert_to_path)
@click.option('-o', '--out_dir',
              type=click.Path(exists=False),
              required=True,
              help='Root directory to store all output files.',
              callback=convert_to_path)
@click.option('-m', '--memory',
              type=click.STRING,
              required=False,
              help='Amount of memory to allocate to job. e.g. "8g". Defaults to 8g.',
              default='8g')
@click.option('--cleanup',
              help='Specify this flag to delete all intermediary files except the resulting FASTA assembly.',
              default=False,
              required=False,
              is_flag=True)
@click.option('--version',
              help='Specify this flag to print the version and exit.',
              is_flag=True,
              is_eager=True,
              callback=print_version,
              expose_value=False)
def assemble(fwd_reads, rev_reads, out_dir, memory, cleanup):
    check_all_dependencies()

    sample_id = get_id(fwd_reads=fwd_reads, rev_reads=rev_reads)
    logging.info(f"Starting ProkaryoteAssembly!")
    if out_dir.exists():
        logging.error(f"ERROR: Output directory {out_dir} already exists. Specify a directory that does not yet exist.")
        quit()
    else:
        out_dir.mkdir(parents=True)

    assembly_pipeline(fwd_reads=fwd_reads,
                      rev_reads=rev_reads,
                      out_dir=out_dir,
                      sample_id=sample_id,
                      memory=memory)

    if cleanup:
        total_cleanup(input_dir=out_dir)
    else:
        basic_cleanup(input_dir=out_dir)


def total_cleanup(input_dir: Path):
    all_files = list(input_dir.glob("*"))
    for f in all_files:
        if ".fasta" not in f.name:
            try:
                os.remove(str(f))
            except:
                pass
    try:
        shutil.rmtree(str(input_dir / 'pilon'))
    except FileNotFoundError:
        pass


def basic_cleanup(input_dir: Path):
    all_files = list(input_dir.glob("*"))
    for f in all_files:
        if ".filtered" in f.name:
            os.remove(str(f))
        elif ".contigs.bam" in f.name:
            os.remove(str(f))
        elif ".contigs.fa" in f.name:
            os.remove(str(f))
    try:
        shutil.rmtree(str(input_dir / 'pilon'))
    except FileNotFoundError:
        pass


def assembly_pipeline(fwd_reads: Path, rev_reads: Path, out_dir: Path, sample_id: str, memory: str = '8g'):
    fwd_reads, rev_reads = call_bbduk(fwd_reads=fwd_reads, rev_reads=rev_reads, out_dir=out_dir)
    fwd_reads, rev_reads = call_tadpole(fwd_reads=fwd_reads, rev_reads=rev_reads, out_dir=out_dir)
    assembly = call_skesa(fwd_reads=fwd_reads, rev_reads=rev_reads, out_dir=out_dir, sample_id=sample_id)
    bamfile = call_bbmap(fwd_reads=fwd_reads, rev_reads=rev_reads, out_dir=out_dir, assembly=assembly)
    polished_assembly = call_pilon(out_dir=out_dir, assembly=assembly, prefix=sample_id, bamfile=bamfile, memory=memory)
    shutil.move(src=str(polished_assembly), dst=str(out_dir / polished_assembly.name.replace(".fasta", ".pilon.fasta")))
    return polished_assembly


def get_id(fwd_reads: Path, rev_reads: Path) -> str:
    element_1 = fwd_reads.name.split("_")[0]
    element_2 = rev_reads.name.split("_")[0]
    if element_1 == element_2:
        logging.info(f"Set Sample ID to {element_1}")
        return element_1


def call_pilon(bamfile: Path, out_dir: Path, assembly: Path, prefix: str, memory: str) -> Path:
    out_dir = out_dir / 'pilon'
    os.makedirs(str(out_dir), exist_ok=True)
    cmd = f"pilon -Xmx{memory} --genome {assembly} --bam {bamfile} --outdir {out_dir} --output {prefix}"
    run_subprocess(cmd)
    polished_assembly = Path(list(out_dir.glob("*.fasta"))[0])
    return polished_assembly


def call_skesa(fwd_reads: Path, rev_reads: Path, out_dir: Path, sample_id: str) -> Path:
    assembly_out = out_dir / Path(sample_id + ".contigs.fa")

    if assembly_out.exists():
        return assembly_out

    cmd = f'skesa --use_paired_ends --gz --fastq "{fwd_reads},{rev_reads}" --contigs_out {assembly_out}'
    run_subprocess(cmd)
    return assembly_out


def index_bamfile(bamfile: Path):
    cmd = f"samtools index {bamfile}"
    run_subprocess(cmd)


def call_bbmap(fwd_reads: Path, rev_reads: Path, out_dir: Path, assembly: Path) -> Path:
    outbam = out_dir / assembly.with_suffix(".bam").name

    if outbam.exists():
        return outbam

    cmd = f"bbmap.sh in1={fwd_reads} in2={rev_reads} ref={assembly} out={outbam} overwrite=t bamscript=bs.sh; sh bs.sh"
    run_subprocess(cmd)

    sorted_bam_file = out_dir / outbam.name.replace(".bam", "_sorted.bam")
    index_bamfile(sorted_bam_file)
    return sorted_bam_file


def call_tadpole(fwd_reads: Path, rev_reads: Path, out_dir: Path) -> tuple:
    fwd_out = out_dir / fwd_reads.name.replace(".filtered.", ".corrected.")
    rev_out = out_dir / rev_reads.name.replace(".filtered.", ".corrected.")

    if fwd_out.exists() and rev_out.exists():
        return fwd_out, rev_out

    cmd = f"tadpole.sh in1={fwd_reads} in2={rev_reads} out1={fwd_out} out2={rev_out} mode=correct"
    run_subprocess(cmd)
    return fwd_out, rev_out


def call_bbduk(fwd_reads: Path, rev_reads: Path, out_dir: Path) -> tuple:
    fwd_out = out_dir / fwd_reads.name.replace(".fastq.gz", ".filtered.fastq.gz")
    rev_out = out_dir / rev_reads.name.replace(".fastq.gz", ".filtered.fastq.gz")

    if fwd_out.exists() and rev_out.exists():
        return fwd_out, rev_out

    cmd = f"bbduk.sh in1={fwd_reads} in2={rev_reads} out1={fwd_out} out2={rev_out} " \
        f"ref=adapters maq=12 qtrim=rl tpe tbo overwrite=t"
    run_subprocess(cmd)
    return fwd_out, rev_out


if __name__ == "__main__":
    assemble()



import logging
import shutil
from pathlib import Path
from subprocess import Popen, PIPE
from ProkaryoteAssembly.config import DEPENDENCIES, __version__, __author__, __email__


def dependency_check(dependency: str) -> bool:
    """
    Checks if a given program is present in the user's $PATH
    :param dependency: String of program name
    :return: True if program is in $PATH, False if not
    """
    check = shutil.which(dependency)
    if check is not None:
        return True
    else:
        return False


def check_all_dependencies():
    # Dependency check
    logging.info("Conducting dependency check...")
    dependency_dict = dict()
    for dependency in DEPENDENCIES:
        dependency_dict[dependency] = dependency_check(dependency)
    if False in dependency_dict.values():
        logging.error("ERROR: Cannot locate some dependencies in $PATH...")
        for key, value in dependency_dict.items():
            if not value:
                logging.error(f"Dependency missing: {key}")
        quit()
    else:
        for key, value in dependency_dict.items():
            logging.debug(f"Dependency {key}: {value}")
    logging.info("Dependencies OK")


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    logging.info(f"Version: {__version__}")
    logging.info(f"Author: {__author__}")
    logging.info(f"Email: {__email__}")
    quit()


def convert_to_path(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    return Path(value)


def run_subprocess(cmd: str, get_stdout=False):
    if get_stdout:
        p = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE)
        out, err = p.communicate()
        out = out.decode().strip()
        err = err.decode().strip()
        if out != "":
            return out
        elif err != "":
            return err
        else:
            return ""
    else:
        p = Popen(cmd, shell=True)
        p.wait()



import setuptools
from ProkaryoteAssembly.config import __email__, __author__, __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

# TODO: Improve this setup by using the layout used in SeqPresenceAbsence
setuptools.setup(
    name="ProkaryoteAssembly",
    install_requires=['click'],
    version=__version__,
    author=__author__,
    author_email=__email__,
    description="Internal BFSSI package for assembling prokaryotic genomes from short reads",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bfssi-forest-dussault/ProkaryoteAssembly",
    packages=setuptools.find_packages(),
    scripts=['ProkaryoteAssembly/prokaryote_assemble.py', 'ProkaryoteAssembly/prokaryote_assemble_dir.py']
)
